# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import asyncio
import re
import traceback

# for concurrent processing
from contextlib import suppress
from typing import Dict, List, Optional

import aiohttp
from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.client.facet_v2 import error_message_run
from openlineage.common.provider.dbt import DbtCloudArtifactProcessor, ParentRunMetadata

from airflow.hooks.base import BaseHook
from airflow.providers.dbt.cloud.hooks.dbt import (
    DbtCloudHook,
    fallback_to_default_account,
)


class DbtCloudExtractor(BaseExtractor):
    default_schema = "public"

    def __init__(self, operator):
        super().__init__(operator)

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return [
            "DbtCloudRunJobOperator",
            "DbtCloudJobRunSensor",
            "DbtCloudRunJobOperatorAsync",
            "DbtCloudJobRunSensorAsync",
        ]

    def get_task_metadata(self):
        try:
            operator = self.operator
            task_name = f"{operator.dag_id}.{operator.task_id}"
            run_facets: Dict = {}
            job_facets: Dict = {}

        except Exception as e:
            self.log.exception("Exception has occurred in extract()")
            error = error_message_run.ErrorMessageRunFacet(str(e), "python")
            error.stackTrace = traceback.format_exc()
            run_facets["errorMessage"] = error
            raise Exception(e)
        finally:
            # finally return the task metadata
            return TaskMetadata(
                name=task_name,
                inputs=[],
                outputs=[],
                run_facets=run_facets,
                job_facets=job_facets,
            )

    def extract(self) -> TaskMetadata:
        operator = self.operator
        # according to the job type, pre-perform the necessary
        # fetch (job and project) of data so that it will save
        # time when creating lineage data at the completion
        if operator.task_type in [
            "DbtCloudRunJobOperator",
            "DbtCloudRunJobOperatorAsync",
        ]:
            job_id = operator.job_id
            hook = DbtCloudHook(operator.dbt_cloud_conn_id)
            # use job_id to get connection
            job = hook.get_job(job_id=job_id).json()["data"]
            project_id = job["project_id"]
            project = hook.get_project(project_id=project_id).json()["data"]
            connection = project["connection"]
            self.context["connection"] = connection
            self.context["job"] = job

        elif operator.task_type in [
            "DbtCloudJobRunSensor",
            "DbtCloudJobRunSensorAsync",
        ]:
            run_id = operator.run_id
            connection = BaseHook.get_connection(operator.dbt_cloud_conn_id)
            account_id = connection.login
            hook = DbtCloudHook(operator.dbt_cloud_conn_id)
            job_run = hook.get_job_run(account_id=account_id, run_id=run_id, include_related=["job"]).json()[
                "data"
            ]
            project_id = job_run["project_id"]
            job = job_run["job"]
            project = hook.get_project(project_id=project_id, account_id=account_id).json()["data"]
            connection = project["connection"]
            self.context["connection"] = connection
            self.context["job"] = job

        return self.get_task_metadata()

    # internal method to extract dbt lineage and send it to
    # OL backend
    def extract_dbt_lineage(self, operator, run_id, job_name=None):
        from openlineage.airflow.adapter import (
            _DAG_NAMESPACE,
            _PRODUCER,
            OpenLineageAdapter,
        )

        job = self.context["job"]
        account_id = job["account_id"]
        hook = DbtCloudHook(operator.dbt_cloud_conn_id)
        execute_steps = job["execute_steps"]
        job_run = hook.get_job_run(
            run_id=run_id, account_id=account_id, include_related=["run_steps"]
        ).json()["data"]
        run_steps = job_run["run_steps"]
        connection = self.context["connection"]
        steps = []

        for run_step in run_steps:
            name = run_step["name"]
            if name.startswith("Invoke dbt with `"):
                regex_pattern = "Invoke dbt with `([^`.]*)`"
                m = re.search(regex_pattern, name)
                command = m.group(1)
                if command in execute_steps:
                    steps.append(run_step["index"])

        headers, tenant = DbtCloudExtractor.get_headers_tenants_from_connection(hook)

        catalog = None
        with suppress(Exception):
            catalog = hook.get_job_run_artifact(run_id, path="catalog.json").json()["data"]

        async def get_steps(steps):
            tasks = [
                self.get_dbt_artifacts(
                    hook,
                    run_id,
                    account_id,
                    step,
                    ["manifest", "run_results"],
                    headers,
                    tenant,
                )
                for step in steps
            ]
            return await asyncio.gather(*tasks)

        loop = asyncio.new_event_loop()
        step_artifacts = loop.run_until_complete(get_steps(steps))
        for artifacts in step_artifacts:
            # process manifest
            manifest = artifacts["manifest"]
            run_reason = manifest["metadata"]["env"]["DBT_CLOUD_RUN_REASON"]
            # ex: Triggered via Apache Airflow by task 'trigger_job_run1' in the astronomy DAG.
            # regex pattern:
            #   Triggered via Apache Airflow by task '[a-zA-Z0-9_]+' in the [a-zA-Z-0-9_]+ DAG.
            regex_pattern = (
                "Triggered via Apache Airflow by task '([a-zA-Z0-9_]+)' " "in the ([a-zA-Z0-9_]+) DAG\\."
            )
            m = re.search(regex_pattern, run_reason)
            task_id = m.group(1)
            dag_id = m.group(2)

            if not artifacts.get("run_results", None):
                continue

            processor = DbtCloudArtifactProcessor(
                producer=_PRODUCER,
                job_namespace=_DAG_NAMESPACE,
                skip_errors=False,
                logger=self.log,
                manifest=manifest,
                run_result=artifacts["run_results"],
                profile=connection,
                catalog=catalog,
            )

            # parent run is an Airflow task triggering DBT Cloud run
            parent_run_id = self.context["task_uuid"]

            # if parent run id exists, set it as parent run metadata - so that
            # the DBT job can contain parent information (which is the current task)
            if parent_run_id is not None:
                parent_job = ParentRunMetadata(
                    run_id=parent_run_id,
                    job_name=job_name if job_name is not None else f"{dag_id}.{task_id}",
                    job_namespace=_DAG_NAMESPACE,
                )
                processor.dbt_run_metadata = parent_job

            events = processor.parse().events()
            client = OpenLineageAdapter.get_or_create_openlineage_client()
            for event in events:
                client.emit(event)

    async def get_dbt_artifacts(self, hook, run_id, account_id, step, artifacts, headers, tenant):
        async with aiohttp.ClientSession(headers=headers) as session:
            tasks = {
                artifact: DbtCloudExtractor.get_job_run_artifact(
                    hook=hook,
                    session=session,
                    tenant=tenant,
                    run_id=run_id,
                    path=artifact,
                    account_id=account_id,
                    step=step,
                )
                for artifact in artifacts
            }
            results = await asyncio.gather(*tasks.values())
        return {filename: result for filename, result in zip(tasks.keys(), results)}

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        task_meta_data = self.get_task_metadata()
        operator = self.operator
        if (
            operator.task_type in ["DbtCloudRunJobOperator", "DbtCloudRunJobOperatorAsync"]
            and operator.wait_for_termination is True
        ):
            run_id = task_instance.xcom_pull(task_ids=task_instance.task_id, key="return_value")
            if not run_id:
                run_url = task_instance.xcom_pull(task_ids=task_instance.task_id, key="job_run_url")
                run_id = run_url.split("/")[-2]
        elif operator.task_type in [
            "DbtCloudJobRunSensor",
            "DbtCloudJobRunSensorAsync",
        ]:
            run_id = operator.run_id

        self.extract_dbt_lineage(
            operator=operator,
            run_id=run_id,
            job_name=f"{task_instance.dag_id}.{task_instance.task_id}",
        )

        return task_meta_data

    @staticmethod
    def get_headers_tenants_from_connection(hook):
        """Get Headers, tenants from the connection details"""
        headers = {}
        connection = hook.get_connection(hook.dbt_cloud_conn_id)
        tenant = connection.host or "cloud.getdbt.com"
        headers["Content-Type"] = "application/json"
        headers["Authorization"] = f"Token {connection.password}"
        return headers, tenant

    @staticmethod
    @fallback_to_default_account
    async def get_job_run_artifact(hook, session, tenant, run_id: int, path: str, account_id=None, step=None):
        endpoint = f"{account_id}/runs/{run_id}/artifacts/{path}.json"
        url = f"https://{tenant}/api/v2/accounts/{endpoint or ''}"
        async with session.get(url, params={"step": step}) as response:
            try:
                response.raise_for_status()
                return await response.json(content_type=None)
            except aiohttp.ClientResponseError as e:
                raise Exception(str(e.status) + ":" + e.message)



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import List
from urllib.parse import urlparse

from openlineage.airflow.extractors.sql_extractor import SqlExtractor
from openlineage.airflow.utils import try_import_from_string


class PostgresExtractor(SqlExtractor):
    _information_schema_columns = [
        "table_schema",
        "table_name",
        "column_name",
        "ordinal_position",
        "udt_name",
    ]
    _is_information_schema_cross_db = False

    # cluster-identifier
    @property
    def dialect(self):
        return "postgres"

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["PostgresOperator"]

    @classmethod
    def get_connection_uri(cls, conn):
        uri = super().get_connection_uri(conn)
        if uri.startswith("postgresql"):
            uri = uri.replace("postgresql", "postgres", 1)
        return uri

    def _get_scheme(self):
        return "postgres"

    def _get_database(self) -> str:
        if self.conn.schema:
            return self.conn.schema
        else:
            parsed = urlparse(self.conn.get_uri())
            return f"{parsed.path}"

    def _get_authority(self) -> str:
        if self.conn.host and self.conn.port:
            return f"{self.conn.host}:{self.conn.port}"
        else:
            parsed = urlparse(self.conn.get_uri())
            return f"{parsed.hostname}:{parsed.port}"

    def _get_hook(self):
        PostgresHook = try_import_from_string("airflow.providers.postgres.hooks.postgres.PostgresHook")
        return PostgresHook(
            postgres_conn_id=self.operator.postgres_conn_id,
            schema=self.operator.database,
        )



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import List, Optional, Type

from openlineage.airflow.extractors import BaseExtractor, Extractors, TaskMetadata
from openlineage.airflow.utils import get_job_name, get_operator_class, get_unknown_source_attribute_run_facet


class ExtractorManager:
    """Class abstracting management of custom extractors."""

    def __init__(self):
        self.extractors = {}
        self.task_to_extractor = Extractors()
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

    def add_extractor(self, operator, extractor: Type[BaseExtractor]):
        self.task_to_extractor.add_extractor(operator, extractor)

    def extract_metadata(
        self, dagrun, task, complete: bool = False, task_instance=None, task_uuid=None
    ) -> TaskMetadata:
        extractor = self._get_extractor(task)
        task_info = (
            f"task_type={get_operator_class(task).__name__} "
            f"airflow_dag_id={task.dag_id} "
            f"task_id={task.task_id} "
            f"airflow_run_id={dagrun.run_id} "
        )

        if extractor:
            # Extracting advanced metadata is only possible when extractor for particular operator
            # is defined. Without it, we can't extract any input or output data.
            if task_uuid:
                extractor.set_context("task_uuid", task_uuid)
            try:
                self.log.debug(f"Using extractor {extractor.__class__.__name__} {task_info}")
                if complete:
                    task_metadata = extractor.extract_on_complete(task_instance)
                else:
                    task_metadata = extractor.extract()

                self.log.debug(f"Found task metadata for operation {task.task_id}: {task_metadata}")
                if task_metadata:
                    if (not task_metadata.inputs) and (not task_metadata.outputs):
                        inlets = task.get_inlet_defs()
                        outlets = task.get_outlet_defs()
                        self.extract_inlets_and_outlets(task_metadata, inlets, outlets)

                    return task_metadata
            except ImportError as e:
                if e.name is not None and e.name.startswith("airflow.providers.openlineage"):
                    self.log.warning(
                        "Failed to import Airflow OpenLineage provider. "
                        "If you're using Airflow 2.7+, please run "
                        "`pip install apache-airflow-providers-openlineage`."
                    )
                else:
                    self.log.exception(
                        f"Failed to extract metadata {e} {task_info}",
                    )
            except Exception as e:
                self.log.exception(
                    f"Failed to extract metadata {e} {task_info}",
                )
        else:
            self.log.debug(f"Unable to find an extractor. {task_info}")

            # Only include the unkonwnSourceAttribute facet if there is no extractor
            task_metadata = TaskMetadata(
                name=get_job_name(task),
                run_facets=get_unknown_source_attribute_run_facet(task=task),
            )
            inlets = task.get_inlet_defs()
            outlets = task.get_outlet_defs()
            self.extract_inlets_and_outlets(task_metadata, inlets, outlets)
            return task_metadata

        return TaskMetadata(name=get_job_name(task))

    def _get_extractor(self, task) -> Optional[BaseExtractor]:
        self.task_to_extractor.instantiate_abstract_extractors(task)
        if task.task_id in self.extractors:
            return self.extractors[task.task_id]
        extractor = self.task_to_extractor.get_extractor_class(get_operator_class(task))
        self.log.debug(f"extractor for {task.__class__} is {extractor}")
        if extractor:
            self.extractors[task.task_id] = extractor(task)
            return self.extractors[task.task_id]
        return None

    def extract_inlets_and_outlets(
        self,
        task_metadata: TaskMetadata,
        inlets: List,
        outlets: List,
    ):
        from openlineage.airflow.extractors.converters import convert_to_dataset

        self.log.debug("Manually extracting lineage metadata from inlets and outlets")
        for i in inlets:
            d = convert_to_dataset(i)
            if d:
                task_metadata.inputs.append(d)
        for o in outlets:
            d = convert_to_dataset(o)
            if d:
                task_metadata.outputs.append(d)



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from typing import Any, Dict, List, Optional

from openlineage.airflow.extractors.base import TaskMetadata
from openlineage.client.facet_v2 import (
    BaseFacet,
    data_quality_assertions_dataset,
    data_quality_metrics_input_dataset,
)


def get_check_extractors(super_):
    class BaseSqlCheckExtractor(super_):
        def __init__(self, operator):
            super().__init__(operator)

        def extract(self) -> TaskMetadata:
            pass

        def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
            from openlineage.airflow.extractors.bigquery_extractor import (
                BigQueryExtractor,
            )

            if issubclass(BigQueryExtractor, BaseSqlCheckExtractor):
                return super().extract_on_complete(task_instance)
            return super().extract()

    class SqlCheckExtractor(BaseSqlCheckExtractor):
        def __init__(self, operator):
            super().__init__(operator)

        @classmethod
        def get_operator_classnames(cls) -> List[str]:
            return ["SQLCheckOperator"]

        def _get_input_facets(self) -> Dict[str, BaseFacet]:
            return {}

    class SqlValueCheckExtractor(BaseSqlCheckExtractor):
        def __init__(self, operator):
            super().__init__(operator)

        @classmethod
        def get_operator_classnames(cls) -> List[str]:
            return ["SQLValueCheckOperator"]

        def _get_input_facets(self) -> Dict[str, BaseFacet]:
            return {}

    class SqlThresholdCheckExtractor(BaseSqlCheckExtractor):
        def __init__(self, operator):
            super().__init__(operator)

        @classmethod
        def get_operator_classnames(cls) -> List[str]:
            return ["SQLThresholdCheckOperator"]

        def _get_input_facets(self) -> Dict[str, BaseFacet]:
            return {}

    class SqlIntervalCheckExtractor(BaseSqlCheckExtractor):
        def __init__(self, operator):
            super().__init__(operator)

        @classmethod
        def get_operator_classnames(cls) -> List[str]:
            return ["SQLIntervalCheckOperator"]

        def _get_input_facets(self) -> Dict[str, BaseFacet]:
            return {}

    class SqlColumnCheckExtractor(BaseSqlCheckExtractor):
        def __init__(self, operator):
            super().__init__(operator)

        @classmethod
        def get_operator_classnames(cls) -> List[str]:
            return ["SQLColumnCheckOperator", "BigQueryColumnCheckOperator"]

        def _get_input_facets(self) -> Dict[str, BaseFacet]:
            """
            Function should expect the column_mapping to take the following form:
            {
                'col_name': {
                    'null_check': {
                        'pass_value': 0,
                        'result': 0,
                        'success': True
                    },
                    'min': {
                        'pass_value': 5,
                        'tolerance': 0.2,
                        'result': 1,
                        'success': False
                    }
                }
            }
            """

            def map_facet_name(check_name) -> str:
                if "null" in check_name:
                    return "nullCount"
                elif "distinct" in check_name:
                    return "distinctCount"
                elif "sum" in check_name:
                    return "sum"
                elif "count" in check_name:
                    return "count"
                elif "min" in check_name:
                    return "min"
                elif "max" in check_name:
                    return "max"
                elif "quantiles" in check_name:
                    return "quantiles"
                return ""

            facet_data: Dict[str, Any] = {"columnMetrics": defaultdict(dict)}
            assertion_data: Dict[str, List[data_quality_assertions_dataset.Assertion]] = {"assertions": []}
            for col_name, checks in self.operator.column_mapping.items():
                col_name = col_name.upper() if self._is_uppercase_names else col_name
                for check, check_values in checks.items():
                    facet_key = map_facet_name(check)
                    facet_data["columnMetrics"][col_name][facet_key] = check_values.get("result")

                    assertion_data["assertions"].append(
                        data_quality_assertions_dataset.Assertion(
                            assertion=check,
                            success=check_values.get("success"),
                            column=col_name,
                        )
                    )
                facet_data["columnMetrics"][col_name] = data_quality_metrics_input_dataset.ColumnMetrics(
                    **facet_data["columnMetrics"][col_name]
                )

            data_quality_facet = data_quality_metrics_input_dataset.DataQualityMetricsInputDatasetFacet(
                **facet_data
            )
            data_quality_assertions_facet = data_quality_assertions_dataset.DataQualityAssertionsDatasetFacet(
                assertions=assertion_data["assertions"]
            )

            return {
                "dataQuality": data_quality_facet,
                "dataQualityMetrics": data_quality_facet,
                "dataQualityAssertions": data_quality_assertions_facet,
            }

    class SqlTableCheckExtractor(BaseSqlCheckExtractor):
        def __init__(self, operator):
            super().__init__(operator)

        @classmethod
        def get_operator_classnames(cls) -> List[str]:
            return ["SQLTableCheckOperator", "BigQueryTableCheckOperator"]

        def _get_input_facets(self) -> Dict[str, BaseFacet]:
            """
            Function should expect to take the checks in the following form:
            {
                'row_count_check': {
                    'pass_value': 100,
                    'tolerance': .05,
                    'result': 101,
                    'success': True
                }
            }
            """
            facet_data: Dict[str, Any] = {"columnMetrics": defaultdict(dict)}
            assertion_data: Dict[str, List[data_quality_assertions_dataset.Assertion]] = {"assertions": []}
            for check, check_values in self.operator.checks.items():
                assertion_data["assertions"].append(
                    data_quality_assertions_dataset.Assertion(
                        assertion=check,
                        success=check_values.get("success"),
                    )
                )
            facet_data["rowCount"] = self.operator.checks.get("row_count_check", {}).get("result", None)
            facet_data["bytes"] = self.operator.checks.get("bytes", {}).get("result", None)

            data_quality_facet = data_quality_metrics_input_dataset.DataQualityMetricsInputDatasetFacet(
                **facet_data
            )
            data_quality_assertions_facet = data_quality_assertions_dataset.DataQualityAssertionsDatasetFacet(
                assertions=assertion_data["assertions"]
            )

            return {
                "dataQuality": data_quality_facet,
                "dataQualityMetrics": data_quality_facet,
                "dataQualityAssertions": data_quality_assertions_facet,
            }

    return [
        SqlCheckExtractor,
        SqlValueCheckExtractor,
        SqlThresholdCheckExtractor,
        SqlIntervalCheckExtractor,
        SqlColumnCheckExtractor,
        SqlTableCheckExtractor,
    ]



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from openlineage.airflow.adapter import _DAG_NAMESPACE, OpenLineageAdapter
from openlineage.airflow.utils import get_job_name

if typing.TYPE_CHECKING:
    from airflow.models import TaskInstance


def lineage_job_namespace():
    """
    Macro function which returns Airflow OpenLineage namespace.
    Invoke as a jinja template, e.g.

    PythonOperator(
        task_id='render_template',
        python_callable=my_task_function,
        op_args=['{{ macros.OpenLineagePlugin.lineage_job_namespace() }}'],
        provide_context=False,
        dag=dag
    )
    """
    return _DAG_NAMESPACE


def lineage_job_name(task_instance: TaskInstance):
    """
    Macro function which returns Airflow task name in OpenLineage format (`<dag_id>.<task_id>`).
    Invoke as a jinja template, e.g.

    PythonOperator(
        task_id='render_template',
        python_callable=my_task_function,
        op_args=['{{ macros.OpenLineagePlugin.lineage_job_name(task_instance) }}'],
        provide_context=False,
        dag=dag
    )
    """
    return get_job_name(task_instance)


def lineage_run_id(task_instance: TaskInstance):
    """
    Macro function which returns the generated runId (UUID) for a given Airflow task. This
    can be used to forward the run id from a task to a child run so the job
    hierarchy is preserved. Invoke as a jinja template, e.g.

    PythonOperator(
        task_id='render_template',
        python_callable=my_task_function,
        op_args=['{{ macros.OpenLineagePlugin.lineage_run_id(task_instance) }}'],
        provide_context=False,
        dag=dag
    )
    """
    return OpenLineageAdapter.build_task_instance_run_id(
        dag_id=task_instance.dag_id,
        task_id=task_instance.task_id,
        try_number=task_instance.try_number,
        execution_date=task_instance.execution_date,
    )


def lineage_parent_id(task_instance: TaskInstance):
    """
    Macro function which returns the job name and runId for a given Airflow task. This
    can be used to forward the ids from a task to a child run so the job
    hierarchy is preserved. Child run can create ParentRunFacet from those ids.

    Result format: `<namespace>/<job_name>/<run_id>`.

    Invoke as a jinja template, e.g.

    PythonOperator(
        task_id='render_template',
        python_callable=my_task_function,
        op_args=['{{ macros.OpenLineagePlugin.lineage_parent_id(task_instance) }}'],
        provide_context=False,
        dag=dag
    )
    """
    return "/".join(
        [
            lineage_job_namespace(),
            lineage_job_name(task_instance),
            lineage_run_id(task_instance),
        ]
    )



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import socket
from ftplib import FTP_PORT
from typing import List, Optional

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.airflow.utils import try_import_from_string
from openlineage.common.dataset import Dataset, Source

FTPOperation = try_import_from_string("airflow.providers.ftp.operators.ftp.FTPOperation")


class FTPExtractor(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["FTPFileTransmitOperator"]

    def extract(self) -> Optional[TaskMetadata]:
        scheme = "file"

        local_host = socket.gethostname()
        try:
            local_host = socket.gethostbyname(local_host)
        except Exception as e:
            self.log.warning(
                f"Failed to resolve local hostname. Using the hostname got by socket.gethostbyname() without resolution. {e}",  # noqa: E501
                exc_info=True,
            )

        conn = self.operator.hook.get_conn()
        remote_host = conn.host
        remote_port = conn.port

        if isinstance(self.operator.local_filepath, str):
            local_filepath = [self.operator.local_filepath]
        else:
            local_filepath = self.operator.local_filepath
        if isinstance(self.operator.remote_filepath, str):
            remote_filepath = [self.operator.remote_filepath]
        else:
            remote_filepath = self.operator.remote_filepath

        local_datasets = [
            Dataset(source=self._get_source(scheme, local_host, None, path), name=path)
            for path in local_filepath
        ]
        remote_datasets = [
            Dataset(
                source=self._get_source(scheme, remote_host, remote_port, path),
                name=path,
            )
            for path in remote_filepath
        ]

        if self.operator.operation.lower() == FTPOperation.GET:
            inputs = remote_datasets
            outputs = local_datasets
        else:
            inputs = local_datasets
            outputs = remote_datasets

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=[ds.to_openlineage_dataset() for ds in inputs],
            outputs=[ds.to_openlineage_dataset() for ds in outputs],
            run_facets={},
            job_facets={},
        )

    def _get_source(self, scheme, host, port, path) -> Source:
        port = port or FTP_PORT
        authority = f"{host}:{port}"
        return Source(
            scheme=scheme,
            authority=authority,
            connection_url=f"{scheme}://{authority}{path}",
        )



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import urlparse

from openlineage.airflow.extractors.base import BaseExtractor, TaskMetadata
from openlineage.airflow.extractors.dbapi_utils import (
    TablesHierarchy,
    create_information_schema_query,
    get_table_schemas,
)
from openlineage.airflow.utils import get_connection
from openlineage.client.facet_v2 import (
    BaseFacet,
    column_lineage_dataset,
    extraction_error_run,
    sql_job,
)
from openlineage.common.dataset import Dataset, Source
from openlineage.common.sql import DbTableMeta, SqlMeta, parse

if TYPE_CHECKING:
    from airflow.hooks.base import BaseHook
    from airflow.models import Connection


class SqlExtractor(BaseExtractor):
    _information_schema_columns = [
        "table_schema",
        "table_name",
        "column_name",
        "ordinal_position",
        "udt_name",
    ]
    _information_schema_table_name = "information_schema.columns"
    _is_information_schema_cross_db = False
    _is_uppercase_names = False
    _allow_trailing_semicolon = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._conn = None
        self._hook = None
        self._database: Optional[str] = None
        self._scheme: Optional[str] = None

    def extract(self) -> TaskMetadata:
        task_name = f"{self.operator.dag_id}.{self.operator.task_id}"
        job_facets = {"sql": sql_job.SQLJobFacet(query=self._normalize_sql(self.operator.sql))}
        run_facets: Dict = {}

        # (1) Parse sql statement to obtain input / output tables.
        self.log.debug(f"Sending SQL to parser: {self.operator.sql}")
        sql_meta: Optional[SqlMeta] = parse(
            self.operator.sql, dialect=self.dialect, default_schema=self.default_schema
        )
        self.log.debug(f"Got meta {sql_meta}")

        if not sql_meta:
            return TaskMetadata(
                name=task_name,
                inputs=[],
                outputs=[],
                run_facets=run_facets,
                job_facets=job_facets,
            )

        if sql_meta.errors:
            run_facets["extractionError"] = extraction_error_run.ExtractionErrorRunFacet(
                totalTasks=len(self.operator.sql) if isinstance(self.operator.sql, list) else 1,
                failedTasks=len(sql_meta.errors),
                errors=[
                    extraction_error_run.Error(
                        errorMessage=error.message,
                        stackTrace=None,
                        task=error.origin_statement,
                        taskNumber=error.index,
                    )
                    for error in sql_meta.errors
                ],
            )

        # (2) Construct source object
        source = Source(
            scheme=self.scheme,
            authority=self._get_authority(),
            connection_url=self.get_connection_uri(self.conn),
        )

        database = getattr(self.operator, "database", None) or self._get_database()

        # (3) Map input / output tables to dataset objects with source set
        # as the current connection. We need to also fetch the schema for the
        # input tables to format the dataset name as:
        # {schema_name}.{table_name}
        in_tables = [t for t in sql_meta.in_tables if not t.provided_field_schema]
        out_tables = [t for t in sql_meta.out_tables if not t.provided_field_schema]
        inputs, outputs = get_table_schemas(
            self.hook,
            source,
            database,
            self._information_schema_query(in_tables) if in_tables else None,
            self._information_schema_query(out_tables) if out_tables else None,
        )

        # (4) Map external stage tables to datasets
        # (no need to query information schema)
        for in_table in sql_meta.in_tables:
            if in_table.provided_field_schema:
                if in_table.provided_namespace:
                    inputs.append(Dataset.from_table(source=source, table_name=in_table.name))
                else:
                    inputs.append(
                        Dataset.from_table(
                            source=source,
                            table_name=in_table.name,
                            schema_name=in_table.schema,
                            database_name=in_table.database or database,
                        )
                    )
        for out_table in sql_meta.out_tables:
            if out_table.provided_field_schema:
                if out_table.provided_namespace:
                    outputs.append(Dataset.from_table(source=source, table_name=out_table.name))
                else:
                    outputs.append(
                        Dataset.from_table(
                            source=source,
                            table_name=out_table.name,
                            schema_name=out_table.schema,
                            database_name=out_table.database or database,
                        )
                    )

        for ds in inputs:
            ds.input_facets = self._get_input_facets()

        for ds in outputs:
            ds.output_facets = self._get_output_facets()
            if len(outputs) == 1:  # Should be always true
                self.attach_column_facet(ds, database, sql_meta)

        db_specific_run_facets = self._get_db_specific_run_facets(source, inputs, outputs)

        run_facets = {**run_facets, **db_specific_run_facets}

        return TaskMetadata(
            name=task_name,
            inputs=[ds.to_openlineage_dataset() for ds in inputs],
            outputs=[ds.to_openlineage_dataset() for ds in outputs],
            run_facets=run_facets,
            job_facets=job_facets,
        )

    def _conn_id(self) -> str:
        if hasattr(self.hook, "conn_id"):
            return "conn_id"
        return getattr(self.hook, self.hook.conn_name_attr)

    @property
    def dialect(self):
        return "generic"

    @property
    def default_schema(self):
        return "public"

    @property
    def hook(self):
        if not self._hook:
            self._hook = SqlExtractor._get_hook(self) or self._get_hook()
        return self._hook

    @property
    def conn(self) -> "Connection":
        if not self._conn:
            self._conn = get_connection(self._conn_id())
        return self._conn  # type: ignore

    @property
    def scheme(self) -> Optional[str]:
        if not self._scheme:
            self._scheme = self._get_scheme()
        return self._scheme

    @property
    def database(self) -> Optional[str]:
        if not self._database:
            self._database = self._get_database()
        return self._database

    @abstractmethod
    def _get_scheme(self) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def _get_database(self) -> Optional[str]:
        raise NotImplementedError

    def _get_authority(self) -> str:
        if self.conn.host and self.conn.port:
            return f"{self.conn.host}:{self.conn.port}"
        else:
            parsed = urlparse(self.conn.get_uri())
            return f"{parsed.hostname}:{parsed.port}"

    def _get_hook(self) -> "Optional[BaseHook]":
        return getattr(self.operator, "_hook", None)

    def _get_db_specific_run_facets(
        self,
        source: Source,
        inputs: Tuple[List[Dataset], ...],
        outputs: Tuple[List[Dataset], ...],
    ) -> Mapping[str, BaseFacet]:
        return {}

    def _get_input_facets(self) -> Dict[str, BaseFacet]:
        return {}

    def _get_output_facets(self) -> Dict[str, BaseFacet]:
        return {}

    @staticmethod
    def _normalize_name(name: str) -> str:
        return name.lower()

    def attach_column_facet(self, dataset: Dataset, database: Optional[str], sql_meta: SqlMeta) -> None:
        if not len(sql_meta.column_lineage):
            return
        dataset.custom_facets["columnLineage"] = column_lineage_dataset.ColumnLineageDatasetFacet(
            fields={
                column_lineage.descendant.name: column_lineage_dataset.Fields(
                    inputFields=[
                        column_lineage_dataset.InputField(
                            namespace=dataset.source.name,
                            name=".".join(
                                filter(
                                    None,
                                    (
                                        column_meta.origin.database or database,
                                        column_meta.origin.schema or self.default_schema,
                                        column_meta.origin.name,
                                    ),
                                )
                            )
                            if column_meta.origin
                            else "",
                            field=column_meta.name,
                        )
                        for column_meta in column_lineage.lineage
                    ],
                    transformationType="",
                    transformationDescription="",
                )
                for column_lineage in sql_meta.column_lineage
            }
        )

    def _information_schema_query(self, tables: List[DbTableMeta]) -> str:
        tables_hierarchy = self._get_tables_hierarchy(
            tables,
            normalize_name=self._normalize_name,
            database=self.database,
            is_cross_db=self._is_information_schema_cross_db,
        )
        return create_information_schema_query(
            columns=self._information_schema_columns,
            information_schema_table_name=self._information_schema_table_name,
            tables_hierarchy=tables_hierarchy,
            uppercase_names=self._is_uppercase_names,
            allow_trailing_semicolon=self._allow_trailing_semicolon,
        )

    @staticmethod
    def _normalize_sql(sql: Union[str, Iterable[str]]):
        if isinstance(sql, str):
            sql = [stmt for stmt in sql.split(";") if stmt != ""]
        sql = [obj for stmt in sql for obj in stmt.split(";") if obj != ""]
        return ";\n".join(sql)

    @staticmethod
    def _get_tables_hierarchy(
        tables: List[DbTableMeta],
        normalize_name: Callable[[str], str],
        database: Optional[str] = None,
        is_cross_db: bool = False,
    ) -> TablesHierarchy:
        hierarchy: TablesHierarchy = {}
        for table in tables:
            if is_cross_db:
                db = table.database or database
            else:
                db = None
            hierarchy.setdefault(normalize_name(db) if db else db, {}).setdefault(
                normalize_name(table.schema) if table.schema else None, []
            ).append(table.name)
        return hierarchy



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import ClassVar, Dict, List

import attr
from openlineage.airflow.version import __version__ as OPENLINEAGE_AIRFLOW_VERSION
from openlineage.client.facet_v2 import BaseFacet
from openlineage.client.utils import RedactMixin

from airflow.version import version as AIRFLOW_VERSION


@attr.s
class AirflowVersionRunFacet(BaseFacet):
    operator: str = attr.ib()
    taskInfo: Dict[str, object] = attr.ib()
    airflowVersion: str = attr.ib()
    openlineageAirflowVersion: str = attr.ib()

    _additional_skip_redact: ClassVar[List[str]] = [
        "operator",
        "airflowVersion",
        "openlineageAirflowVersion",
    ]

    @classmethod
    def from_dagrun_and_task(cls, dagrun, task):
        from openlineage.airflow.utils import get_operator_class

        return cls(
            operator=f"{get_operator_class(task).__module__}.{get_operator_class(task).__name__}",
            taskInfo={},
            airflowVersion=AIRFLOW_VERSION,
            openlineageAirflowVersion=OPENLINEAGE_AIRFLOW_VERSION,
        )


@attr.s
class AirflowRunArgsRunFacet(BaseFacet):
    externalTrigger: bool = attr.ib(default=False)

    _additional_skip_redact: ClassVar[List[str]] = ["externalTrigger"]


@attr.s
class AirflowMappedTaskRunFacet(BaseFacet):
    mapIndex: int = attr.ib()
    operatorClass: str = attr.ib()

    _additional_skip_redact: ClassVar[List[str]] = ["operatorClass"]

    @classmethod
    def from_task_instance(cls, task_instance):
        task = task_instance.task
        from openlineage.airflow.utils import get_operator_class

        return cls(
            task_instance.map_index,
            f"{get_operator_class(task).__module__}.{get_operator_class(task).__name__}",
        )


@attr.s
class AirflowRunFacet(BaseFacet):
    """
    Composite Airflow run facet.
    """

    dag: Dict = attr.ib()
    dagRun: Dict = attr.ib()
    task: Dict = attr.ib()
    taskInstance: Dict = attr.ib()
    taskUuid: str = attr.ib()


@attr.s
class UnknownOperatorInstance(RedactMixin):
    """
    Describes an unknown operator - specifies the (class) name of the operator
    and its properties
    """

    name: str = attr.ib()
    properties: Dict[str, object] = attr.ib()
    type: str = attr.ib(default="operator")

    _skip_redact: ClassVar[List[str]] = ["name", "type"]


@attr.s
class UnknownOperatorAttributeRunFacet(BaseFacet):
    """
    RunFacet that describes unknown operators in an Airflow DAG
    """

    unknownItems: List[UnknownOperatorInstance] = attr.ib()



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import List
from urllib.parse import urlparse

from openlineage.airflow.extractors.sql_extractor import SqlExtractor
from openlineage.airflow.utils import try_import_from_string


class TrinoExtractor(SqlExtractor):
    _information_schema_columns = [
        "table_schema",
        "table_name",
        "column_name",
        "ordinal_position",
        "data_type",
        "table_catalog",
    ]
    _is_information_schema_cross_db = True
    _allow_trailing_semicolon = False

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return ["TrinoOperator"]

    @property
    def default_schema(self):
        return self.conn.schema

    def _get_scheme(self):
        return "trino"

    def _get_database(self) -> str:
        # hive is default in airflow trino provider, not in trino package
        return self.conn.extra_dejson.get("catalog", "hive")

    def _get_authority(self) -> str:
        if self.conn.host and self.conn.port:
            return f"{self.conn.host}:{self.conn.port}"
        else:
            parsed = urlparse(self.conn.get_uri())
            return f"{parsed.hostname}:{parsed.port}"

    def _get_hook(self):
        TrinoHook = try_import_from_string("airflow.providers.trino.hooks.trino.TrinoHook")
        return TrinoHook(trino_conn_id=self.operator.trino_conn_id)



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
import re
import sys
from unittest.mock import patch

import pytest
from openlineage.airflow.utils import get_location

from tests.mocks.git_mock import execute_git_mock

log = logging.getLogger(__name__)


@patch("openlineage.airflow.utils.execute_git", side_effect=execute_git_mock)
def test_dag_location(git_mock):
    assert re.match(
        r"https://github.com/[^/]+/OpenLineage/blob/"
        "abcd1234/integration/airflow/tests/test_dags/test_dag.py",
        get_location("tests/test_dags/test_dag.py"),
    )


@patch("openlineage.airflow.utils.execute_git", side_effect=execute_git_mock)
def test_bad_file_path(git_mock):
    log.debug("test_bad_file_path()")
    with pytest.raises(FileNotFoundError):
        # invalid file
        get_location("dags/missing-dag.py")


if __name__ == "__main__":
    pytest.main([sys.argv[0]])



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
import os
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional

from openlineage.airflow.extractors import TaskMetadata
from openlineage.airflow.utils import DagUtils, redact_with_exclusions
from openlineage.airflow.version import __version__ as OPENLINEAGE_AIRFLOW_VERSION
from openlineage.client import OpenLineageClient, OpenLineageClientOptions, set_producer
from openlineage.client.event_v2 import Job, Run, RunEvent, RunState
from openlineage.client.facet_v2 import (
    JobFacet,
    RunFacet,
    documentation_job,
    error_message_run,
    job_type_job,
    nominal_time_run,
    ownership_job,
    parent_run,
    processing_engine_run,
    source_code_location_job,
)
from openlineage.client.uuid import generate_static_uuid

from airflow.stats import Stats

if TYPE_CHECKING:
    from airflow.models.dagrun import DagRun


_DAG_DEFAULT_OWNER = "anonymous"
_DAG_DEFAULT_NAMESPACE = "default"

_DAG_NAMESPACE = os.getenv("OPENLINEAGE_NAMESPACE", os.getenv("MARQUEZ_NAMESPACE", _DAG_DEFAULT_NAMESPACE))

_PRODUCER = (
    f"https://github.com/OpenLineage/OpenLineage/tree/{OPENLINEAGE_AIRFLOW_VERSION}/integration/airflow"
)

set_producer(_PRODUCER)

# https://openlineage.io/docs/spec/facets/job-facets/job-type
# They must be set after the `set_producer(_PRODUCER)`
# otherwise the `JobTypeJobFacet._producer` will be set with the default value
_JOB_TYPE_DAG = job_type_job.JobTypeJobFacet(jobType="DAG", integration="AIRFLOW", processingType="BATCH")
_JOB_TYPE_TASK = job_type_job.JobTypeJobFacet(jobType="TASK", integration="AIRFLOW", processingType="BATCH")

log = logging.getLogger(__name__)


class OpenLineageAdapter:
    """
    Adapter for translating Airflow metadata to OpenLineage events,
    instead of directly creating them from Airflow code.
    """

    _client = None

    def __init__(self):
        if "OPENLINEAGE_AIRFLOW_LOGGING" in os.environ:
            logging.getLogger(__name__.rpartition(".")[0]).setLevel(os.getenv("OPENLINEAGE_AIRFLOW_LOGGING"))

    @staticmethod
    def get_or_create_openlineage_client() -> OpenLineageClient:
        # Backcomp with Marquez integration
        marquez_url = os.getenv("MARQUEZ_URL")
        if marquez_url:
            log.info(f"Sending lineage events to {marquez_url}")
            client = OpenLineageClient(
                marquez_url,
                OpenLineageClientOptions(api_key=os.environ["MARQUEZ_API_KEY"]),
            )
        else:
            client = OpenLineageClient()
        return client

    @property
    def client(self) -> OpenLineageClient:
        if not self._client:
            self._client = OpenLineageAdapter.get_or_create_openlineage_client()
        return self._client

    @staticmethod
    def build_dag_run_id(dag_id: str, execution_date: datetime) -> str:
        return str(
            generate_static_uuid(
                instant=execution_date,
                data=f"{_DAG_NAMESPACE}.{dag_id}".encode("utf-8"),
            )
        )

    @staticmethod
    def build_task_instance_run_id(
        dag_id: str,
        task_id: str,
        try_number: int,
        execution_date: datetime,
    ) -> str:
        return str(
            generate_static_uuid(
                instant=execution_date,
                data=f"{_DAG_NAMESPACE}.{dag_id}.{task_id}.{try_number}".encode("utf-8"),
            )
        )

    def emit(self, event: RunEvent):
        event = redact_with_exclusions(event)
        try:
            with Stats.timer("ol.emit.attempts"):
                return self.client.emit(event)
        except Exception as e:
            Stats.incr("ol.emit.failed")
            log.exception(f"Failed to emit OpenLineage event of id {event.run.runId}")
            log.debug(e)

    def start_task(
        self,
        run_id: str,
        job_name: str,
        job_description: str,
        event_time: str,
        parent_job_name: Optional[str],
        parent_run_id: Optional[str],
        code_location: Optional[str],
        nominal_start_time: str,
        nominal_end_time: str,
        owners: List[str],
        task: Optional[TaskMetadata],
        run_facets: Optional[Dict[str, RunFacet]] = None,  # Custom run facets
    ) -> str:
        """
        Emits openlineage event of type START
        :param run_id: globally unique identifier of task in dag run
        :param job_name: globally unique identifier of task in dag
        :param job_description: user provided description of job
        :param event_time:
        :param parent_job_name: the name of the parent job (typically the DAG,
                but possibly a task group)
        :param parent_run_id: identifier of job spawning this task
        :param code_location: file path or URL of DAG file
        :param nominal_start_time: scheduled time of dag run
        :param nominal_end_time: following schedule of dag run
        :param owners: list of owners of DAG
        :param task: metadata container with information extracted from operator
        :param run_facets: custom run facets
        :return:
        """
        from airflow.version import version as AIRFLOW_VERSION

        processing_engine_version_facet = processing_engine_run.ProcessingEngineRunFacet(
            version=AIRFLOW_VERSION,
            name="Airflow",
            openlineageAdapterVersion=OPENLINEAGE_AIRFLOW_VERSION,
        )
        if run_facets is None:
            run_facets = {}
        if task:
            run_facets = {**task.run_facets, **run_facets}
        run_facets["processing_engine"] = processing_engine_version_facet  # type: ignore
        event = RunEvent(
            eventType=RunState.START,
            eventTime=event_time,
            run=self._build_run(
                run_id=run_id,
                parent_job_name=parent_job_name,
                parent_run_id=parent_run_id,
                job_name=job_name,
                nominal_start_time=nominal_start_time,
                nominal_end_time=nominal_end_time,
                run_facets=run_facets,
            ),
            job=self._build_job(
                job_name=job_name,
                job_description=job_description,
                code_location=code_location,
                owners=owners,
                job_facets=task.job_facets if task else None,
                job_type=_JOB_TYPE_TASK,
            ),
            inputs=task.inputs if task else [],
            outputs=task.outputs if task else [],
        )
        self.emit(event)
        return event.run.runId

    def complete_task(
        self,
        run_id: str,
        parent_job_name: Optional[str],
        parent_run_id: Optional[str],
        job_name: str,
        end_time: str,
        task: TaskMetadata,
    ):
        """
        Emits openlineage event of type COMPLETE
        :param run_id: globally unique identifier of task in dag run
        :param job_name: globally unique identifier of task between dags
        :param parent_job_name: the name of the parent job (typically the DAG,
                but possibly a task group)
        :param parent_run_id: identifier of job spawning this task
        :param end_time: time of task completion
        :param task: metadata container with information extracted from operator
        """

        event = RunEvent(
            eventType=RunState.COMPLETE,
            eventTime=end_time,
            run=self._build_run(
                run_id=run_id,
                parent_job_name=parent_job_name,
                parent_run_id=parent_run_id,
                run_facets=task.run_facets,
            ),
            job=self._build_job(job_name, job_facets=task.job_facets, job_type=_JOB_TYPE_TASK),
            inputs=task.inputs,
            outputs=task.outputs,
        )
        self.emit(event)

    def fail_task(
        self,
        run_id: str,
        job_name: str,
        parent_job_name: Optional[str],
        parent_run_id: Optional[str],
        end_time: str,
        task: TaskMetadata,
    ):
        """
        Emits openlineage event of type FAIL
        :param run_id: globally unique identifier of task in dag run
        :param job_name: globally unique identifier of task between dags
        :param parent_job_name: the name of the parent job (typically the DAG,
                but possibly a task group)
        :param parent_run_id: identifier of job spawning this task
        :param end_time: time of task completion
        :param task: metadata container with information extracted from operator
        """
        event = RunEvent(
            eventType=RunState.FAIL,
            eventTime=end_time,
            run=self._build_run(
                run_id=run_id,
                parent_job_name=parent_job_name,
                parent_run_id=parent_run_id,
                run_facets=task.run_facets,
            ),
            job=self._build_job(job_name, job_facets=task.job_facets, job_type=_JOB_TYPE_TASK),
            inputs=task.inputs,
            outputs=task.outputs,
        )
        self.emit(event)

    def dag_started(
        self,
        dag_run: "DagRun",
        msg: str,
        nominal_start_time: str,
        nominal_end_time: str,
    ):
        event = RunEvent(
            eventType=RunState.START,
            eventTime=DagUtils.to_iso_8601(dag_run.start_date),
            job=Job(name=dag_run.dag_id, namespace=_DAG_NAMESPACE, facets={"jobType": _JOB_TYPE_DAG}),
            run=self._build_run(
                run_id=self.build_dag_run_id(
                    dag_id=dag_run.dag_id,
                    execution_date=dag_run.execution_date,
                ),
                nominal_start_time=nominal_start_time,
                nominal_end_time=nominal_end_time,
            ),
            inputs=[],
            outputs=[],
        )
        self.emit(event)

    def dag_success(self, dag_run: "DagRun", msg: str):
        event = RunEvent(
            eventType=RunState.COMPLETE,
            eventTime=DagUtils.to_iso_8601(dag_run.end_date),
            job=Job(name=dag_run.dag_id, namespace=_DAG_NAMESPACE, facets={"jobType": _JOB_TYPE_DAG}),
            run=Run(
                runId=self.build_dag_run_id(
                    dag_id=dag_run.dag_id,
                    execution_date=dag_run.execution_date,
                ),
            ),
            inputs=[],
            outputs=[],
        )
        self.emit(event)

    def dag_failed(self, dag_run: "DagRun", msg: str):
        event = RunEvent(
            eventType=RunState.FAIL,
            eventTime=DagUtils.to_iso_8601(dag_run.end_date),
            job=Job(name=dag_run.dag_id, namespace=_DAG_NAMESPACE, facets={"jobType": _JOB_TYPE_DAG}),
            run=Run(
                runId=self.build_dag_run_id(dag_id=dag_run.dag_id, execution_date=dag_run.execution_date),
                facets={
                    "errorMessage": error_message_run.ErrorMessageRunFacet(
                        message=msg, programmingLanguage="python"
                    )
                },
            ),
            inputs=[],
            outputs=[],
        )
        self.emit(event)

    @staticmethod
    def _build_run(
        run_id: str,
        parent_job_name: Optional[str] = None,
        parent_run_id: Optional[str] = None,
        job_name: Optional[str] = None,
        nominal_start_time: Optional[str] = None,
        nominal_end_time: Optional[str] = None,
        run_facets: Dict[str, RunFacet] = None,
    ) -> Run:
        facets: Dict[str, RunFacet] = {}
        if nominal_start_time:
            facets.update(
                {"nominalTime": nominal_time_run.NominalTimeRunFacet(nominal_start_time, nominal_end_time)}
            )
        parent_name = parent_job_name or job_name
        if parent_run_id is not None and parent_name is not None:
            parent_run_facet = parent_run.ParentRunFacet(
                run=parent_run.Run(runId=parent_run_id),
                job=parent_run.Job(namespace=_DAG_NAMESPACE, name=parent_name),
            )
            facets.update(
                {
                    "parent": parent_run_facet,
                }
            )

        if run_facets:
            facets.update(run_facets)

        return Run(run_id, facets)

    @staticmethod
    def _build_job(
        job_name: str,
        job_type: job_type_job.JobTypeJobFacet,
        job_description: Optional[str] = None,
        code_location: Optional[str] = None,
        owners: List[str] = None,
        job_facets: Dict[str, JobFacet] = None,
    ):
        facets: Dict[str, JobFacet] = {}

        if job_description:
            facets.update({"documentation": documentation_job.DocumentationJobFacet(job_description)})
        if code_location:
            facets.update(
                {"sourceCodeLocation": source_code_location_job.SourceCodeLocationJobFacet("", code_location)}
            )
        if owners:
            facets.update(
                {
                    "ownership": ownership_job.OwnershipJobFacet(
                        owners=[ownership_job.Owner(name=owner) for owner in owners]
                    )
                }
            )
        if job_facets:
            facets = {**facets, **job_facets}

        facets.update({"jobType": job_type})

        return Job(_DAG_NAMESPACE, job_name, facets)
